import numpy as np
import os
import numpy as np
import pandas as pd
import yaml
from tqdm import tqdm
from scipy.optimize import curve_fit
from moleculekit.molecule import Molecule
from moleculekit.projections.metricrmsd import MetricRmsd

import matplotlib.pyplot as plt
import seaborn as sns
sns.set_style("whitegrid")


kB = 0.0019872041 # kcal/mol/K

# Input:  counts and bin limits (n+1)
# Output: counts normalized by spheric shell volumes, and bin centers (n)
def renorm(counts, bins):
    h = bins[1]-bins[0]
    R = .5*(bins[1:]+bins[:-1])  # bin centers
    vols = 4*np.pi/3*(bins[1:]**3-bins[:-1]**3)
    ncounts = counts / vols
    return np.vstack([R,ncounts])


# Bonded fit

def harmonic(x,x0,k,V0):
    return k*(x-x0)**2+V0

# Coarse Graining repulsion fit

def CG(r,eps,V0):
    sigma=1.0
    sigma_over_r = sigma/r
    V = 4*eps*(sigma_over_r**6) + V0
    return V

def get_param_bonded(mol, bond_range, Temp):
    bonds_types = {}
    for bond in mol.bonds:
        btype = tuple(mol.atomtype[bond])
        if btype in bonds_types:
            bonds_types[btype].append(bond)
        elif tuple([btype[1], btype[0]]) in bonds_types:
            bonds_types[tuple([btype[1], btype[0]])].append(bond)
        else:
            bonds_types[btype] = [bond]

    prior_bond = {}

    for bond in bonds_types.keys():
        dists = []
        for idx0, idx1 in bonds_types[bond]:
            dists.append(np.linalg.norm(mol.coords[idx0,:,:] - mol.coords[idx1,:,:], axis=0))

        dist = np.concatenate(dists, axis=0)

        yb, bins= np.histogram(dist, bins=40,  range=bond_range)
        RR, ncounts = renorm(yb, bins)

        # Drop zero counts
        RR_nz = RR[ncounts>0]
        ncounts_nz = ncounts[ncounts>0]
        dG_nz = -kB*Temp*np.log(ncounts_nz)


        # Fit may fail, better to try-catch. p0 usually not necessary if function is reasonable.
        popt, _ = curve_fit(harmonic, RR_nz, dG_nz, p0=[np.array(bond_range).mean(), 60, -1])

        # Just a hard-coded example, the full code requires more changes
        bname=f"({bond[0]}, {bond[1]})"
        prior_bond[bname]={'req': popt[0].tolist(),
                            'k0':  popt[1].tolist() }


        # popt now has the function parameters
        plt.plot(RR_nz, dG_nz, 'o')
        plt.plot(RR_nz, harmonic(RR_nz, *popt))
        plt.title(f'{bond[0]}-{bond[1]}')
        plt.show()

    return prior_bond


def get_param_nonbonded(mol, fit_range, Temp):
    atom_types = {}
    for at in set(mol.atomtype):
        atom_types[at] = np.where(mol.atomtype == at)[0]

    prior_lj = {}

    for at in atom_types.keys():
        dists = []
        for idx in atom_types[at]:
            bonded = []
            for bond in mol.bonds:
                if idx in bond:
                    bonded.append(bond[0])
                    bonded.append(bond[1])

            for idx2 in list(set(mol.bonds.flatten())):
                if idx2 not in bonded:
                    dists.append(np.linalg.norm(mol.coords[idx, :, :] - mol.coords[idx2, :, :], axis=0))

        dist = np.concatenate(dists, axis=0)
        # save only below 6A
        dist = dist[dist < fit_range[1]]

        yb, bins = np.histogram(dist, bins=30, range=fit_range)  ### Adjust the range if needed
        RR, ncounts = renorm(yb, bins)

        RR_nz = RR[ncounts > 0]
        ncounts_nz = ncounts[ncounts > 0]
        dG_nz = -kB * Temp * np.log(ncounts_nz)

        popt, _ = curve_fit(CG, RR_nz, dG_nz, p0=[2229.0, 1.])

        # Just a hard-coded example, the full code requires more changes
        bname = at
        prior_lj[bname] = {'epsilon': popt[0].tolist(),
                           'sigma': 1.0}

        plt.plot(RR_nz, dG_nz, 'o')
        plt.plot(RR_nz, CG(RR_nz, *popt))
        plt.title(f'{at}')
        plt.show()

    return prior_lj


import torch
import numpy as np
from tqdm import tqdm
from torchmd.forcefields.forcefield import ForceField
from torchmd.forces import Forces
from torchmd.systems import System
from moleculekit.molecule import Molecule
from torchmd.parameters import Parameters


def make_deltaforces(
    coords_npz,
    forces_npz,
    delta_forces_npz,
    forcefield,
    psf,
    exclusions=("bonds"),
    device="cpu",
    forceterms=["Bonds", "Angles", "RepulsionCG"],
):
    device = torch.device(device)
    precision = torch.double
    replicas = 1

    mol = Molecule(psf)
    atom_types = mol.atomtype
    natoms = mol.numAtoms

    coords = np.load(coords_npz)

    ### test if any bonds are longer than 10A
    print("Check for broken coords.")

    broken_frames = []
    for bond in tqdm(mol.bonds):
        crds = coords[:, bond, :]
        crds_dif = crds[:, 0, :] - crds[:, 1, :]
        dists = np.linalg.norm(crds_dif, axis=1)
        broken = dists > 10.0
        broken_frames.append(broken)

    broken_frames = np.stack(broken_frames)
    broken_frames = broken_frames.any(axis=0)

    if broken_frames.any():
        print("Removing broken coords with distances larger than 10A.")

        coords_good = coords[~broken_frames, :, :]  # remove broken frames
        np.save(coords_npz.replace(".", "_fix."), coords_good)

        broken_coords = coords[broken_frames, :, :]
        np.save(coords_npz.replace(".", "_broken."), broken_coords)

        coords = coords_good

    else:
        print("No broken frames")

    all_forces = np.load(forces_npz)[~broken_frames, :, :]
    coords = torch.tensor(coords, dtype=precision).to(device)

    atom_vel = torch.zeros(replicas, natoms, 3)
    atom_forces = torch.zeros(natoms, 3, replicas)
    atom_pos = torch.zeros(natoms, 3, replicas)
    box_full = torch.zeros(3, replicas)

    ff = ForceField.create(mol, forcefield)
    parameters = Parameters(ff, mol, forceterms, precision=precision, device=device)

    system = System(natoms, replicas, precision, device)
    system.set_positions(atom_pos)
    system.set_box(box_full)
    system.set_velocities(atom_vel)

    forces = Forces(parameters, terms=forceterms, exclusions=exclusions)

    print("Producing delta forces")
    prior_forces = []
    for co in tqdm(coords):
        Epot = forces.compute(co.reshape([1, natoms, 3]), system.box, system.forces)
        fr = (
            system.forces.detach().cpu().numpy().astype(np.float32).reshape([natoms, 3])
        )
        prior_forces.append(fr)

    prior_forces = np.array(prior_forces)
    delta_forces = all_forces - prior_forces

    np.save(delta_forces_npz, delta_forces)



import numpy as np
from moleculekit.molecule import Molecule
from torchmd_cg.utils.mappings import CA_MAP, CACB_MAP


def pdb2psf_CA(pdb_name_in, psf_name_out, bonds=True, angles=True):
    mol = Molecule(pdb_name_in)
    mol.filter("name CA")

    n = mol.numAtoms

    atom_types = []
    for i in range(n):
        atom_types.append(CA_MAP[(mol.resname[i], mol.name[i])])

    if bonds:
        bonds = np.concatenate(
            (
                np.arange(n - 1).reshape([n - 1, 1]),
                (np.arange(1, n).reshape([n - 1, 1])),
            ),
            axis=1,
        )
    else:
        bonds = np.empty([0, 2], dtype=np.int32)

    if angles:
        angles = np.concatenate(
            (
                np.arange(n - 2).reshape([n - 2, 1]),
                (np.arange(1, n - 1).reshape([n - 2, 1])),
                (np.arange(2, n).reshape([n - 2, 1])),
            ),
            axis=1,
        )
    else:
        angles = np.empty([0, 3], dtype=np.int32)

    mol.atomtype = np.array(atom_types)
    mol.bonds = bonds
    mol.angles = angles
    mol.write(psf_name_out)


def pdb2psf_CACB(pdb_name_in, psf_name_out, bonds=True, angles=True):
    mol = Molecule(pdb_name_in)
    mol.filter("name CA CB")

    n = mol.numAtoms

    atom_types = []
    for i in range(n):
        atom_types.append(CACB_MAP[(mol.resname[i], mol.name[i])])

    CA_idx = []
    CB_idx = []
    for i, name in enumerate(mol.name):
        if name[:2] == "CA":
            CA_idx.append(i)
        else:
            CB_idx.append(i)

    if bonds:
        CA_bonds = np.concatenate(
            (
                np.array(CA_idx)[:-1].reshape([len(CA_idx) - 1, 1]),
                np.array(CA_idx)[1:].reshape([len(CA_idx) - 1, 1]),
            ),
            axis=1,
        )
        CB_bonds = np.concatenate(
            (
                np.array(CB_idx).reshape(len(CB_idx), 1) - 1,
                np.array(CB_idx).reshape(len(CB_idx), 1),
            ),
            axis=1,
        )
        bonds = np.concatenate((CA_bonds, CB_bonds))
    else:
        bonds = np.empty([0, 2], dtype=np.int32)

    if angles:
        CA_angles = np.concatenate(
            (
                np.array(CA_idx)[:-2].reshape([len(CA_idx) - 2, 1]),
                np.array(CA_idx)[1:-1].reshape([len(CA_idx) - 2, 1]),
                np.array(CA_idx)[2:].reshape([len(CA_idx) - 2, 1]),
            ),
            axis=1,
        )

        CB_angles = []
        cbn = 0
        for i in CA_idx:
            if mol.resname[i] != "GLY":
                if i != 0:
                    CB_angles.append([CA_idx[CA_idx.index(i) - 1], i, CB_idx[cbn]])
                if i != CA_idx[-1]:
                    CB_angles.append([CA_idx[CA_idx.index(i) + 1], i, CB_idx[cbn]])
                cbn += 1
        CB_angles = np.array(CB_angles)
        angles = np.concatenate((CA_angles, CB_angles))

    else:
        angles = np.empty([0, 3], dtype=np.int32)

    mol.atomtype = np.array(atom_types)
    mol.bonds = bonds
    mol.angles = angles
    mol.write(psf_name_out)
